/* FINAL PROJECT
	PWM with
    IMU with ITG-3200 Gyro and ADXL345 Accelerometer Test Code
	Nov 2013
		
	Use complementary filter
	Tested on a 5V 16MHz ATMega1284 Breakout Board
	10kOhm pull-ups on I2C lines. No pull-ups on SPI lines

	ADXL345:
	SCL -> SCK  (B7)
	SDA -> MOSI (B5)
	SDO -> MISO (B6)
	CS  -> SS	(B4)
	INT -> no connection, not used here

	ITG-3200:
	VDD & VIO = 5V
	SDA -> SDA  (C1)
	SCL -> SCL	(C0)
	INT -> no connection, not used here
	CLK -> GND
*/

#include "trtSettings.h"
#include "trtkernel_1284.c"
#include <stdio.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include <stdlib.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <math.h>
#include "i2c.h"
#include "ITG-3200.h"

/* CPU frequency */
#define F_CPU 16000000UL

//Gyro and Accelerometer
#define ITG3200_R 0xD3	// ADD pin is pulled high
#define ITG3200_W 0xD2	// So address is 0x69

#define twoKpDef  (2.0f * 0.5f) // 2 * proportional gain
#define twoKiDef  (2.0f * 0.1f) // 2 * integral gain

#define sbi(var, mask)   ((var) |= (uint8_t)(1 << mask))
#define cbi(var, mask)   ((var) &= (uint8_t)~(1 << mask))

#define adxl PINB4 // ADXL345

// defines for complimentary filter
#define RateConstant 2.0f
#define dt 0.03f

//defines used to allow for more manageable numbers for PID
#define MAX_INTEGRAL 5000
#define SCALE 30

// ISR
#define SUSPEND cli();
#define RESUME sei();

// serial communication library
// Don't mess with the semaphores
#define SEM_RX_ISR_SIGNAL 1
#define SEM_STRING_DONE 2 // user hit <enter>
#include "trtUart.h"
#include "trtUart.c"

// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

// semaphore to protect shared variable
#define SEM_SHARED 3

// the usual
#define begin {
#define end }

#define t1 2000
volatile unsigned int time1;

// shorthand for setting pin values
#define READ(U, N) ((U) >> (N) & 1u)
#define SET(U, N) ((void)((U) |= 1u << (N)))
#define CLR(U, N) ((void)((U) &= ~(1u << (N))))
#define FLIP(U, N) ((void)((U) ^= 1u << (N))) 

// --- control LEDs from buttons and uart -------------------
// input arguments to each thread
// not actually used in this example
int args[2] ;

// shared PWM and PID parameters
uint8_t led ;
float xangle;
float yangle;
float prop;
float integ;
float deriv;
volatile int motor_period;
volatile int motor_period_ovlf;
volatile unsigned int angle;
float sensor_output;

///============Function Prototypes for Gyroscope=========/////////////////
char ITG3200Read(unsigned char address);
void ITG3200Write(unsigned char address, unsigned char data);
void ITG3200ViewRegisters(void);
int checkInterrupt(void);
void getITG3200(int average,long *_GyroX, long *_GyroY, long *_GyroZ);
void ITG3200BlockRead(unsigned char address);

///============General Use Prototypes=====//////////////////
void ioinit(void);

void iniSPI(void);
void send_byte(uint8_t byte);

// functions for determining angles for position of platform
float comp_filter(float previousAngle, float newAngle, float newRate);
void readGyro(float *_GyroX, float *_GyroY, float *_GyroZ);
float invSqrt(float number);
void getValues(float *values);
void getAccEuler(float *angles);

volatile unsigned long timer0_overflow_count; 
volatile uint16_t time;

float angles[4];
float gyro[3];
float filter_angle[2];
float filterTerm2 = 0;

// --- timer 0 interrupt -------------------------------------
ISR(TIMER0_OVF_vect) 
{ 
	++timer0_overflow_count; 
	if (time>0)	--time;
} 

// --- define task 1  ----------------------------------------
void PC(void* args) 
begin
	double inputNum ;
	char cmd[4];
	
	while(1)
	begin
		// commands:
		// 'x 80' 	angle for pitch is 80 degrees from neutral position 
		// 'y -40'  angle for roll is -40 degrees from neutral position
		// 'p 10' 	proportional is 10
		// 'i 2' 	integral is 2
		// 'd 5.2' 	derivative is 5.2
		fprintf(stdout, ">") ;
		fscanf(stdin, "%s%f", cmd, &inputNum) ;
		trtWait(SEM_STRING_DONE);

		trtWait(SEM_SHARED) ;
		if (cmd[0] == 'x')
			xangle = inputNum ;
		if (cmd[0] == 'y')
			yangle = inputNum ;
		if (cmd[0] == 'p')		
			prop = inputNum ;
		if (cmd[0] == 'i')		
			integ  = inputNum ;
		if (cmd[0] == 'd')		
			deriv  = inputNum ;
		trtSignal(SEM_SHARED);
		if (cmd[0] == 's')	
		{	
			printflag = 1 ;
			while(!(UCSR0A & (1<<RXC0))){_delay_ms(100);};
			printflag = 0;
		}
		fprintf(stdout, "%s == %f\n\r", cmd, inputNum) ;
		fprintf(stdout, "x == %f\n\r", (double)xangle) ;
		fprintf(stdout, "y == %f\n\r", (double)yangle) ;
		fprintf(stdout, "p == %f\n\r", (double)prop) ;
		fprintf(stdout, "i == %f\n\r", (double)integ) ;
		fprintf(stdout, "d == %f\n\r", (double)deriv) ;
	end
end
  
// --- define task 2  ----------------------------------------
void PID(void* args) 
begin
	uint32_t rel, dead ;
	float x_error,y_error;
	float x_error_prev = 0;
	float y_error_prev = 0;
	float m_xinput,m_yinput;
	float pid_xintegral = 0;
	float pid_yintegral = 0;
	float x_move = 1500;
	float y_move = 1500;

	while(1)
	begin

		trtWait(SEM_SHARED) ;
		
		//obtain angles determined from complimentary filter
		xangle = filter_angle[0];
		yangle = -filter_angle[1];
	
		x_error = xangle; 
		y_error = yangle; 
		
		trtSignal(SEM_SHARED);
		
		//set max limit on errors to prevent overshoot
		if (x_error>70) x_error = 70;
		if (y_error>70) y_error = 70;

		// if statements ultimately not really used - attempted to make position change only in response to large movements
		if (((x_error>0)-(x_error<0)) == ((x_error_prev>0)-(x_error_prev<0)))
			pid_xintegral += x_error;
		else 
			pid_xintegral = 0;
  
		if (((y_error>0)-(y_error<0)) == ((y_error_prev>0)-(y_error_prev<0)))
			pid_yintegral += y_error;
		else 
			pid_yintegral = 0;
		
		//Standard PID calculations from lab 4
		if (pid_xintegral > MAX_INTEGRAL) pid_xintegral = MAX_INTEGRAL;
 		if (pid_yintegral > MAX_INTEGRAL) pid_yintegral = MAX_INTEGRAL;
		 
		m_xinput = prop * x_error + integ*pid_xintegral / 1000 + deriv*(x_error - x_error_prev);
		m_yinput = prop * y_error + integ*pid_yintegral / 1000 + deriv*(y_error - y_error_prev);

		// converting angles to corresponding values for the PWM
		// -90 degrees to +90 degrees correspond to OCR values of 770 to 2230 
		x_move += (m_xinput*8.1111)/SCALE;
		y_move += (m_yinput*8.1111)/SCALE;

		if (x_move>2230) x_move = 2230;
		if (y_move>2230) y_move = 2230;
		if (x_move<770) x_move = 770;
		if (y_move<770) y_move = 770;

		// set values for the PWM to control motor positions
		OCR1A = (int)x_move;
		OCR1B = (int)y_move;
	  	
		x_error_prev = x_error;
		y_error_prev = y_error;	
  		
		//standard sleep
		rel = trtCurrentTime() + SECONDS2TICKS(0.001);
	    dead = trtCurrentTime() + SECONDS2TICKS(0.002);
	    trtSleepUntil(rel, dead);
	end
end

// --- define task 3  ----------------------------------------
void SENSOR(void* args) 
begin
	uint32_t rel, dead ;
	
	while(1)
	begin
		trtWait(SEM_SHARED) ;
		
		//obtain readings from gyroscope and accelerometer
		getAccEuler(&angles);
		readGyro(&gyro[0],&gyro[1],&gyro[2]);

		// determine final angle values after using complimentary filter and gyro and accelerometer data
		filter_angle[0] = comp_filter(filter_angle[0],angles[0],-gyro[0]);
		filter_angle[1] = comp_filter(filter_angle[1],angles[1],gyro[1]);

		trtSignal(SEM_SHARED);

		PORTA ^= 0b00000001;  //used for debugging to see how long this angle calculations take by viewing blinking LED
		
		rel = trtCurrentTime() + SECONDS2TICKS(0.001);
	    dead = trtCurrentTime() + SECONDS2TICKS(0.002);
	    trtSleepUntil(rel, dead);
	end
end

// --- Main Program ----------------------------------
int main(void) {
   
   //------initialize accelerometer and gyro--------------------
   	time = 200;
	uint8_t byte;
//	int8_t data[6];

	trt_uart_init();
	ioinit();
	i2cInit();
	iniSPI();

	_delay_ms(40000);//long _delay
	
	printf("\nInitialize Gyro...\n");  //indicate gyroscope is working

	ITG3200Write(PWR_M, 0x80);	// Reset to defaults
	ITG3200Write(SMPL, 0x00);	// SMLPRT_DIV = 0
	ITG3200Write(DLPF, 0x18);	// DLPF_CFG = 0, FS_SEL = 3
	ITG3200Write(INT_C, 0x05);	// Generate interrupt when device is ready or raw data ready
	ITG3200Write(PWR_M, 0x00);
	
	_delay_ms(1000);

	//waits for power register to respond correctly
	while(byte != 0x08)
	{
		cbi(PORTB, adxl);
		send_byte(0x2D);
		send_byte(0x08);
		sbi(PORTB, adxl);
		
		_delay_ms(5);
		
		cbi(PORTB, adxl);
		send_byte(0xAD);
		send_byte(0x00);
		byte = SPDR;
		sbi(PORTB, adxl);
		
		printf("Power is: %x\n", byte);
		
		_delay_ms(1000);
	}

	cbi(PORTB, adxl);
	send_byte(0x31);
	send_byte(0x01);
	sbi(PORTB, adxl);
	
	_delay_ms(5);
	
	cbi(PORTB, adxl);
	send_byte(0xB1);
	send_byte(0x00);
	byte = SPDR;
	sbi(PORTB, adxl);
	
	printf("Format is: %x\n", byte);
	
	cbi(PORTB, adxl);
	send_byte(0x38);
	send_byte(0x00);
	sbi(PORTB, adxl);
	
	_delay_ms(5);
	
	cbi(PORTB, adxl);
	send_byte(0xB8);
	send_byte(0x00);
	byte = SPDR;
	sbi(PORTB, adxl);
	
	printf("FIFO is: %x\n", byte);
		
	cbi(PORTB, adxl);
	send_byte(0x2C);
	send_byte(0x0F);
	sbi(PORTB, adxl);

	fprintf(stdout,"\nStreaming...\n\n");

	getAccEuler(&angles);
	filter_angle[0] = angles[0];
	filter_angle[1] = angles[1];
	//--------end gyro and accelerometer
   
    //unsigned int angle;
	DDRA = 0b00000001;
	DDRD = (1<<PIND4) | (1<<PIND5) | (1<<PIND2);
	CLR(PORTD,2);
	//USE TIMER 3 for MOTOR CONTROL
	// 8 prescaler to slow down the PWM 
	TCCR1B = 2 | (1<<WGM13); // divide by 8  and set to Phase and Frequency Correct
	// turn on PWM
	// turn on fast PWM and OC3A and OC3B outputs
	TCCR1A = (1<<COM1A1) | (1<<COM1B1) ;
	OCR1A = 1500 ; // controls actual movement of servo
	OCR1B = 1500 ;
	ICR1 = 20000; // frequency = fclk/(2*prescaler*ICR1) = 50Hz
	//can change OCR1A between 1000 and 2000 to obtain 1-2ms high pulses

	sei() ; 	// turn on all ISRs
   
	// start TRT
	trtInitKernel(80); // 80 bytes for the idle task stack

	// --- create semaphores ----------
	trtCreateSemaphore(SEM_RX_ISR_SIGNAL, 0) ; // uart receive ISR semaphore
	trtCreateSemaphore(SEM_STRING_DONE,0) ;  // user typed <enter>
  
	// variable protection
	trtCreateSemaphore(SEM_SHARED, 1) ; // protect shared variables

	// --- creat tasks  ----------------
	trtCreateTask(PC, 200, SECONDS2TICKS(0.05), SECONDS2TICKS(0.15), &(args[0]));      // for user control through UART
	trtCreateTask(PID, 300, SECONDS2TICKS(0.001), SECONDS2TICKS(0.002), &(args[1]));   // for motor control
    trtCreateTask(SENSOR, 300, SECONDS2TICKS(0.01), SECONDS2TICKS(0.02), &(args[1]));  // for determining angles
	// --- Idle task --------------------------------------
	// just sleeps the cpu to save power 
	// every time it executes
	set_sleep_mode(SLEEP_MODE_IDLE);
	sleep_enable();
	
	while (1) 
	begin
		sleep_cpu();
	end

} // main

//complimentary filter used to combine gyro and acceleromter measurements for better accuracy
float comp_filter(float previousAngle, float newAngle, float newRate) 
{
	float filterAngle;

	float filterTerm0;
	float filterTerm1;
	float timeConstant;

	timeConstant=1.0; // default 1.0

	//see complimentary filter algorithm in report
	filterTerm0 = (newAngle - previousAngle) * timeConstant * timeConstant;
	filterTerm2 += filterTerm0 * dt *(abs(newRate)/10) ;
	filterTerm1 = filterTerm2 + ((newAngle - previousAngle) * 2 * timeConstant) + newRate*RateConstant;
	filterAngle = (filterTerm1 * dt) + previousAngle;

	return filterAngle; // This is actually the current angle, but is stored for the next iteration
}


void readAccel(int *x, int *y, int *z) 
{
	int8_t data[6];
	//read the acceleration data from the ADXL345
	cbi(PORTB, adxl);
	send_byte(0xF2);

	for(uint8_t i=0; i<6; i++)
	{
		send_byte(0x00);
		data[i] = SPDR;
	}

	sbi(PORTB, adxl);

	// each axis reading comes in 10 bit resolution, ie 2 bytes.  Least Significat Byte first!!
	// thus we are converting both bytes in to one int
	*x = ((data[1] & 0x1F) <<8) | data[0];
	*y = ((data[3] & 0x1F) <<8) | data[2];
	*z = ((data[5] & 0x1F) <<8) | data[4];
}

void readGyro(float *_GyroX, float *_GyroY, float *_GyroZ) 
{
	long x, y, z;
	int average=2 ;

	getITG3200(average, &x, &y, &z); // x,y,z will contain calibrated integer values from the sensor
	*_GyroX =  (float)x / 14.375 + 1.67;
	*_GyroY =  (float)y / 14.375 - 1.67;
	*_GyroZ =  (float)z / 14.375 + 0.49;
}

void getValues(float *values) 
{  
	int accval[3];
	//float gyroval[3];

	readAccel(&accval[0], &accval[1], &accval[2]);
	values[0] = ((float) accval[0]);
	values[1] = ((float) accval[1]);
	values[2] = ((float) accval[2]);

	readGyro(&values[3],&values[4],&values[5]);

}

void getAccEuler(float *angles) 
{
	int accval[3];
	float x_acc_g,y_acc_g,z_acc_g, x_tilt,y_tilt,factor_yz,factor_xz;
	int average=4 ;

	angles[2] = 0;
	angles[3] = 0;

	for (int i=0;i<average;i++)
	{
		readAccel(&accval[0], &accval[1], &accval[2]);
		//getValues(&accval);

		x_acc_g = accval[0] * 0.0078;
		y_acc_g = accval[1] * 0.0078;
		z_acc_g = accval[2] * 0.0078;

		factor_yz = sqrt(y_acc_g * y_acc_g + z_acc_g * z_acc_g);
		factor_xz = sqrt(x_acc_g * x_acc_g + z_acc_g * z_acc_g);

		x_tilt = atan(x_acc_g/factor_yz) * 180 / M_PI ;
		y_tilt = atan(y_acc_g/factor_xz) * 180 / M_PI ;

		angles[0] = x_tilt;
		angles[1] = y_tilt;

		angles[2] += angles[0];
		angles[3] += angles[1];
	}
	angles[0] = angles[2] / average;
	angles[1] = angles[3] / average;
}


/***Gyro and Accelerometer Functions****************************/

char ITG3200Read(unsigned char address)
{
	char data;
	
	cbi(TWCR, TWEN);	// Disable TWI
	sbi(TWCR, TWEN);	// Enable TWI
	
	i2cSendStart();
	i2cWaitForComplete();
	
	i2cSendByte(ITG3200_W);	// write 0xD2
	i2cWaitForComplete();
	
	i2cSendByte(address);	// write register address
	i2cWaitForComplete();
	
	i2cSendStart();
	
	i2cSendByte(ITG3200_R);	// write 0xD3
	i2cWaitForComplete();
	i2cReceiveByte(FALSE);
	i2cWaitForComplete();
	
	data = i2cGetReceivedByte();	// Get MSB result
	i2cWaitForComplete();
	i2cSendStop();
	
	cbi(TWCR, TWEN);	// Disable TWI
	sbi(TWCR, TWEN);	// Enable TWI
	
	return data;
}

void ITG3200Write(unsigned char address, unsigned char data)
{
	i2cSendStart();
	i2cWaitForComplete();
	
	i2cSendByte(ITG3200_W);	// write 0xB4
	i2cWaitForComplete();
	
	i2cSendByte(address);	// write register address
	i2cWaitForComplete();
	
	i2cSendByte(data);
	i2cWaitForComplete();
	
	i2cSendStop();
}

void ITG3200ViewRegisters(void)
{
	printf("\nWHO_AM_I (0x00): 0x%x\n", ITG3200Read(WHO));
	_delay_ms(1000);
	printf("SMPLRT_DIV (0x15): 0x%x\n", ITG3200Read(SMPL));
	printf("DLPF_FS (0x16): 0x%x\n", ITG3200Read(DLPF));
	printf("INT_CFG (0x17): 0x%x\n", ITG3200Read(INT_C));
	printf("INT_STATUS (0x1A): 0x%x\n", ITG3200Read(INT_S));
	printf("TEMP_OUT_H (0x1B): 0x%x\n", ITG3200Read(TMP_H));
	printf("TEMP_OUT_L (0x1C): 0x%x\n", ITG3200Read(TMP_L));
	printf("GYRO_XOUT_H (0x1D): 0x%x\n", ITG3200Read(GX_H));
	printf("GYRO_XOUT_L (0x1E): 0x%x\n", ITG3200Read(GX_L));
	printf("GYRO_YOUT_H (0x1F): 0x%x\n", ITG3200Read(GY_H));
	printf("GYRO_YOUT_L (0x20): 0x%x\n", ITG3200Read(GY_L));
	printf("GYRO_ZOUT_H (0x21): 0x%x\n", ITG3200Read(GZ_H));
	printf("GYRO_ZOUT_L (0x22): 0x%x\n", ITG3200Read(GZ_L));
	printf("PWR_MGM (0x3E): 0x%x\n", ITG3200Read(PWR_M));
}

int checkInterrupt(void)
{
	if ((PIND & (1<<2)) == 0)
		return 0;
	else
		return 1;
}

void getITG3200(int average, long *_GyroX, long *_GyroY, long *_GyroZ)
{
	char temp;
	int gx[average], gy[average], gz[average];
	//int gx = 0;
	//int gy = 0;
	//int gz = 0;
	long gyrox = 0;
	long gyroy = 0;
	long gyroz = 0;
	
	for (int i = 0; i<average; i++)
	{
		while (!(ITG3200Read(INT_S) & 0x01));
		temp = 0;
		temp = ITG3200Read(GX_H);
		gx[i] = temp << 8;
		gx[i] |= ITG3200Read(GX_L);

		//while (!(ITG3200Read(INT_S) & 0x01));
		temp = 0;
		temp = ITG3200Read(GY_H);
		gy[i] = temp << 8;
		gy[i] |= ITG3200Read(GY_L);

		//while (!(ITG3200Read(INT_S) & 0x01));
		temp = 0;
		temp = ITG3200Read(GZ_H);
		gz[i] = temp << 8;
		gz[i] |= ITG3200Read(GZ_L);
		
		//while (!(ITG3200Read(INT_S) & 0x01));
		gyrox = gyrox + gx[i];
		gyroy = gyroy + gy[i];
		gyroz = gyroz + gz[i];
	}
	
	//gyrox = gyrox/average;
	//gyroy = gyroy/average;
	//gyroz = gyroz/average;
	
	*_GyroX = gyrox/average;
	*_GyroY = gyroy/average;
	*_GyroZ = gyroz/average;
	//if (time==0) {printf("	%ld	%ld	%ld\n", gyrox, gyroy, gyroz);}
}

void send_byte(uint8_t byte)
{
	SPDR = byte;
	while(!(SPSR & (1<<SPIF)));
}


float invSqrt(float number) {
  volatile long i;
  volatile float x, y;
  volatile const float f = 1.5F;

  x = number * 0.5F;
  y = number;
  i = *(long*) &y;
  i = 0x5f375a86 - ( i >> 1 );
  y = *(float*) &i;
  y = y * ( f - ( x * y * y ) );
  return y;
}

/*******************Initialize********************/
 //initialize for I2C for gyroscope
void ioinit (void)
{
    //1 = output, 0 = input
    DDRB = 0b00000000; //All inputs
    DDRC = 0b00000011; //PORTC1 (SDA), PORTC0 (SCL)
    //DDRD = 0b000000000; //PORTD (RX on PD0)
	PORTC = 0b00000011; //pullups on the I2C bus
	//sbi(PORTD, 2);

	TCCR0B = 3 ;  // set prescalar to divide by 64
	TIMSK0 = (1<<TOIE0) ;//turn on timer 0 overflow ISR

//	uart_init();
	stdout = stdin = stderr = &uart_str;
	fprintf(stdout,"\nStarting System...\n\r");

	sei();
	
	xangle = 0;
	yangle = 0;
	
	prop = 0;
	deriv = 0;
	integ = 0;
}
//initialize for SPI for accelerometer
void iniSPI(void)
{
	// Set MOSI, SCK, and SS output, all others input 
	DDRB |= (1<<PINB4)|(1<<PINB5) | (1<<PINB7);	 // Set MOSI , SCK , and SS output
	
	//sbi(PORTB, adxl);
	
	SPCR = ( (1<<SPE)|(1<<MSTR) | (1<<SPR1) |(1<<SPR0) |(1<<CPOL) |(1<<CPHA));	// Enable SPI, Master, set clock rate fck/128  

	//sbi(SPSR, 0);
    _delay_ms(1);
}
